import traceback
from typing import Any, NamedTuple, Optional
import graphviz
import torch
import torch.fx


class TensorMetadata(NamedTuple):
    # TensorMetadata is a structure containing pertinent information
    # about a tensor within a PyTorch program.

    # General Tensor metadata
    shape: torch.Size
    dtype: torch.dtype
    requires_grad: bool
    memory_format: Optional[torch.memory_format]

    def __repr__(self):
        return "×".join(map(str, self.shape))


def _extract_tensor_metadata(result: torch.Tensor) -> TensorMetadata:
    """
    Extract a TensorMetadata NamedTuple describing `result`.
    """
    shape = result.shape
    dtype = result.dtype
    requires_grad = result.requires_grad

    memory_formats = {
        torch.contiguous_format,
        torch.channels_last,
        torch.channels_last_3d,
    }

    memory_format = None

    for query_format in memory_formats:
        if result.is_contiguous(memory_format=query_format):
            memory_format = query_format
            break

    return TensorMetadata(shape, dtype, requires_grad, memory_format)


class ResultProbe(torch.fx.Interpreter):
    def run_node(self, n: torch.fx.Node) -> Any:
        try:
            result = super().run_node(n)
        except Exception:
            traceback.print_exc()
            raise RuntimeError(
                f"ShapeProp error for: "
                f"node={n.format_node()} with "
                f"meta={n.meta}"
            )
        find_tensor_in_result = False

        def extract_tensor_meta(obj):
            if isinstance(obj, torch.Tensor):
                nonlocal find_tensor_in_result
                find_tensor_in_result = True
                return _extract_tensor_metadata(obj)
            else:
                return obj

        n.meta["result"] = torch.fx.node.map_aggregate(result, extract_tensor_meta)
        n.meta["find_tensor_in_result"] = find_tensor_in_result
        return result


def html_table(*content, **kwargs):
    kwargs_pairs = [f'{k}="{v}"' for k, v in kwargs.items()]
    return f'<table {" ".join(kwargs_pairs)}>' + "\n".join(content) + "</table>"


def html_tr(*content, **kwargs):
    kwargs_pairs = [f'{k}="{v}"' for k, v in kwargs.items()]
    return f'<tr {" ".join(kwargs_pairs)}>' + "\n".join(content) + "</tr>"


def html_td(content, **kwargs):
    kwargs_pairs = [f'{k}="{v}"' for k, v in kwargs.items()]
    return f'<td {" ".join(kwargs_pairs)}>' + str(content) + "</td>"


def node_label_html(model, node):
    name = node._pretty_print_target(node.target)
    result = node.meta["result"]

    cols = [[html_td(result)]]

    if node.op == "call_module":
        module = model.get_submodule(node.target)
        head = str(module)
        cols[0] = [html_td(name, rowspan=len(cols)), *cols[0]]
    elif node.op == "call_method":
        head = f".{name}()"
    elif node.op == "get_attr":
        head = f".{name}"
    elif node.op == "call_function":
        head = f"{name}()"
    else:
        head = name

    head_kwargs = dict(colspan=len(cols[0]))
    if not node.meta["find_tensor_in_result"]:
        head_kwargs["bgcolor"] = "lightgray"

    html = html_table(
        html_tr(html_td(head, **head_kwargs)),
        *[html_tr(*c) for c in cols],
        border=0,
        cellborder=1,
        cellspacing=0,
    )
    return f"<{html}>"


def single_node(model: torch.nn.Module, graph: graphviz.Digraph, node: torch.fx.Node):
    node_label = node_label_html(model, node)
    node_kwargs = dict(shape="plaintext")
    graph.node(node.name, node_label, **node_kwargs)
    for in_node in node.all_input_nodes:
        edge_kwargs = dict()
        if (
                not node.meta["find_tensor_in_result"]
                or not in_node.meta["find_tensor_in_result"]
        ):
            edge_kwargs.update(dict(style="dashed", color="lightgrey"))
        graph.edge(in_node.name, node.name, **edge_kwargs)


def model_graph(model: torch.nn.Module, *args, **kwargs) -> graphviz.Digraph:
    symbolic_traced: torch.fx.GraphModule = torch.fx.symbolic_trace(model)
    ResultProbe(symbolic_traced).run(*args, **kwargs)
    symbolic_traced.graph.print_tabular()
    graph = graphviz.Digraph("model", format="svg", node_attr={"shape": "plaintext"})
    for node in symbolic_traced.graph.nodes:
        single_node(model, graph, node)
    return graph


def draw(model: torch.nn.Module, inputs: torch.Tensor, save_dir: str = './Save', save_name: str = 'model'):
    graph = model_graph(model, inputs)
    graph.render(outfile=save_dir + '/' + save_name, view=False)


if __name__ == '__main__':
    pass



import torch
from torch import nn
from torchvision.models import resnet18,regnet_x_8gf


class TestConv2d(nn.Module):
    def __init__(self, in_channels, out_channels, **kwargs):
        super(TestConv2d, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, **kwargs)
        self.bn = nn.BatchNorm2d(out_channels)
        self.relu = nn.ReLU(True)

    def forward(self, x):
        x = self.conv(x)
        x = self.bn(x)
        x = self.relu(x)
        return x


class TestModel(nn.Module):
    def __init__(self):
        super().__init__()
        self.conv1 = TestConv2d(3, 32, kernel_size=3)
        self.conv2 = TestConv2d(32, 64, kernel_size=3)
        self.dropout = nn.Dropout(0.3)

    def forward(self, x):
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.dropout(x)
        return x


model= resnet18()
model2=regnet_x_8gf()


import argparse
from utils import draw
import importlib
import torch
from torch import nn
import os


def parse_py(py_path: str) -> list[nn.Module]:
    spec = importlib.util.spec_from_file_location('module_name', py_path)
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    module_classes = [m for _, m in module.__dict__.items() if
                      isinstance(m, (nn.Module, type)) or issubclass(type(m),nn.Module)]
    if len(module_classes) == 0:
        raise ValueError('No nn.Module class found in the file.')
    return module_classes


def main():
    parser = argparse.ArgumentParser(description="Draw a model structure diagram")
    parser.add_argument('--file', '-f', type=str, required=True, help='the file of model(only support .py)')
    parser.add_argument('--input-size', '-i', type=int, nargs='+', default=[224, 224],
                        help='the size of model input,just img size')
    parser.add_argument('--dir', '-d', default='./Save', type=str, help='The path of save dir')
    parser.add_argument('--name', '-n', default='model.svg', type=str, help='The name of save svg')
    args = parser.parse_args()

    if not os.path.exists(args.dir):
        os.makedirs(args.dir)

    model_file = args.file
    model_list = []
    inputs_shape = [1, 3]
    inputs_shape.extend(args.input_size)
    inputs = torch.randn(inputs_shape)

    if model_file.endswith('.py'):
        model_list = parse_py(model_file)
    else:
        print("Just support .py !!!")

    for index, model_name in enumerate(model_list):

        try:
            if not isinstance(model_name, nn.Module):
                model = model_name()

            else:
                model = model_name
            args.name=type(model).__name__+'.svg'
            draw(model, inputs, save_dir=args.dir, save_name=args.name)
        except:
            print(f"{model_name} draw failed")
            pass


if __name__ == "__main__":
    main()
